
# built-in
from contextlib import contextmanager
from collections import ChainMap, abc
from functools import partial
import typing

# packages
from anytree import NodeMixin
from anytree.search import find_by_attr


class Singleton(object):
    
    def __init__(self, callback: typing.Callable[[], typing.Any]):
        self._callback = callback
        self._created = False
        self._value = None
    
    def __call__(self) -> typing.Any:
        if not self._created:
            self._value = self._callback()
            self._created = True
        
        return self._value


class Binder(abc.Mapping):
    
    _registry: typing.MutableMapping[typing.Hashable, typing.Callable[[], typing.Any]]
    
    def __init__(self, registry: typing.MutableMapping=None):
        if registry is None:
            self._registry = {}
        else:
            self._registry = registry
    
    def bind_to_instance(self, key: typing.Hashable, value: typing.Any):
        """
        Store an instance value that won't change.
        """
        self._registry[key] = lambda: value
    
    def bind_to_constructor(self, key: typing.Hashable, value: typing.Callable[[], typing.Any]):
        """
        Store a callable that won't be executed until it's used,
        but always returns the same value thereafter.
        """
        self._registry[key] = Singleton(value)
    
    def bind_to_provider(self, key: typing.Hashable, value: typing.Callable[[], typing.Any]):
        """
        Store a callable that won't be executed until it's used,
        but is executed every time, potentially returning a new value.
        """
        self._registry[key] = value
    
    def clear(self):
        self._registry.clear()
    
    def __iter__(self):
        return iter(self._registry)
    
    def __getitem__(self, item) -> typing.Any:
        return self._registry[item]()
        
    def __contains__(self, item):
        return item in self._registry

    def __repr__(self):
        return self._registry.__repr__()
    
    def __len__(self):
        return len(self._registry)


class AttributeInjection(object):
    
    def __init__(self, value_func: typing.Callable[[], typing.Any]):
        self._value_func = value_func

    def __get__(self, obj, owner) -> typing.Any:
        return self._value_func()


class Injector(NodeMixin):
    
    # 'Injector' is a forward reference to be resolved later (PEP 484)
    def __init__(self, name: str, parent: 'Injector'=None, binder: Binder=None):
        self.name = name
        self.parent = parent
        self._binder = Binder()
    
    def get_sub_injector(self, name: str) -> 'Injector':
        return Injector(name, parent=self)
        
    def configure(self, config: typing.Union[typing.Mapping, typing.Callable[[Binder], None]]):
        if callable(config):
            config(self._binder)
        elif isinstance(config, dict):
            for key, value in config.items():
                self._binder.bind_to_instance(key, value)
        else:
            raise NotImplementedError()
    
    def clear(self):
        self._binder.clear()
    
    def get_instance(self, key) -> typing.Any:
        parent_binders = [a._binder for a in self.ancestors]
        resolver = ChainMap(self._binder, *reversed(parent_binders))

        try:
            value = resolver[key]
        except KeyError:
            raise RuntimeError(f'No binding was found for key={key}')
        
        return value
    
    def get_attribute(self, key, *, scope=None) -> AttributeInjection:
        return AttributeInjection(partial(self.get_instance, key))
        
    def __repr__(self):
        return f"Injector('{self.name}')"
        

@contextmanager
def override(target_injector: Injector):
    old_binder = target_injector._binder
    
    target_injector._binder = Binder(registry=ChainMap({}, old_binder))
    
    try:
        yield target_injector
    finally:
        target_injector._binder = old_binder


injector = Injector('__global__')
